/*
 * Access various peripherals on the MIPS Creator CI20 board.
 *
 * Copyright (C) 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "common.h"
#include "defs.h"

#include <l4/devices/dma.h>
#include <l4/devices/memory.h>

#include <l4/io/io.h>
#include <l4/re/env.h>
#include <l4/re/c/util/cap_alloc.h>
#include <l4/sys/cache.h>
#include <l4/sys/factory.h>
#include <l4/sys/icu.h>
#include <l4/sys/ipc.h>
#include <l4/sys/irq.h>
#include <l4/sys/rcv_endpoint.h>
#include <l4/vbus/vbus.h>

#include <stdio.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>



/* Common configuration. */

static l4_cap_idx_t icucap;

/* Memory definitions. */

static const unsigned int num_memory_regions = 8;

static struct memory_region memory_regions[] = {
  {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0},
  };



/* Device and resource discovery. */

static long item_in_range(long start, long end, long index)
{
  if (start < end)
    return start + index;
  else if (start > end)
    return start - index;
  else
    return start;
}



/* Common functions. */

static int init_irq(int num, l4_cap_idx_t irq, l4_uint32_t start, l4_uint32_t end)
{
  /* Create interrupt object. */

  long err = l4_error(l4_factory_create_irq(l4re_global_env->factory, irq));

  if (err)
  {
    printf("Could not create IRQ object: %ld\n", err);
    return 1;
  }

  /* Bind interrupt objects to IRQ numbers. */

  err = l4_error(l4_icu_bind(icucap,
                             item_in_range(start, end, num),
                             irq));

  if (err)
  {
    printf("Could not bind IRQ to the ICU: %ld\n", err);
    return 1;
  }

  /* Attach ourselves to the interrupt handler. */

  err = l4_error(l4_rcv_ep_bind_thread(irq, l4re_env()->main_thread, num));

  if (err)
  {
    printf("Could not attach to IRQs: %ld\n", err);
    return 1;
  }

  return 0;
}

static const char *delimiters = " \n";

static char *read_token(const char *message)
{
  char *token;

  if ((token = strtok(NULL, delimiters)) == NULL)
  {
    if (message != NULL)
      printf("%s?\n", message);
  }

  return token;
}

static int read_flag(const char *message, int *value, const char *flag, size_t n)
{
  char *token = read_token(message);

  if (token == NULL)
    return 0;

  *value = !strncmp(token, flag, n);
  return 1;
}

static int read_encoded_number(const char *message, const char *format, unsigned int *num)
{
  char *token = read_token(message);

  if (token == NULL)
    return 0;

  sscanf(token, format, num);
  return 1;
}

static int read_number(const char *message, unsigned int *num)
{
  char *token = read_token(message);

  if (token == NULL)
    return 0;

  if (!strncmp(token, "0x", 2) || !strncmp(token, "0X", 2))
    sscanf(token, "%x", num);
  else
    *num = atoi(token);

  return 1;
}

static int get_resource_number(const char *type, unsigned int available)
{
  unsigned int num;

  if (!read_number(type, &num))
    return -1;

  if (num >= available)
  {
    printf("%s number out of range.\n", type);
    return -1;
  }

  return num;
}

static int get_channel_number(unsigned int available)
{
  return get_resource_number("Channel", available);
}

static int get_region_number(unsigned int available)
{
  return get_resource_number("Region", available);
}

static void *get_channel(int num_channels, void *channels[], int *num)
{
  int n = get_channel_number(num_channels);
  void *channel;

  if (n < 0)
    return NULL;

  channel = channels[n];

  if (channel == NULL)
    printf("Channel needs creating.\n");

  if (num != NULL)
    *num = n;

  return channel;
}

static void list_channels(int num_channels, void *channels[])
{
  int num;
  void *channel;

  for (num = 0; num < num_channels; num++)
  {
    printf("Channel %d: ", num);

    channel = channels[num];

    if (channel == NULL)
      printf("(inactive)\n");
    else
      printf("active\n");
  }
}

static struct dma_region *_get_dma_region(void)
{
  int num = get_region_number(num_dma_regions);

  if (num < 0)
    return NULL;

  return &dma_regions[num];
}

static struct memory_region *_get_memory_region(void)
{
  int num = get_region_number(num_memory_regions);

  if (num < 0)
    return NULL;

  return &memory_regions[num];
}

static int get_memory_region(unsigned int size, l4_addr_t *addr)
{
  *addr = (l4_addr_t) calloc(size, sizeof(char));

  return !(*addr);
}



/* AIC/I2S operations. */

static void new_aic_channel(void *aic)
{
  int num = get_channel_number(num_aic_channels);
  void *channel;

  if (num < 0)
    return;

  channel = get_channel(num_dma_channels, dma_channels, NULL);

  if (channel == NULL)
    return;

  aic_channels[num] = aic_get_channel(aic, num, channel);
}

static void aic_transfer_data(void)
{
  void *channel = get_channel(num_aic_channels, aic_channels, NULL);
  struct dma_region *region;
  uint32_t count, rate, resolution, transferred;

  if (channel == NULL)
    return;

  region = _get_dma_region();

  if (region == NULL)
    return;

  if (!read_number("Sample size", &count))
    return;

  if (!read_number("Sample rate", &rate))
    return;

  if (!read_number("Sample resolution", &resolution))
    return;

  transferred = aic_transfer(channel, region->paddr, count, rate, resolution);

  printf("Transferred: %d\n", transferred);
}



/* CPM operations. */

static const char *clock_id(enum Clock_identifiers clock)
{
  for (int i = 0; clocks[i].id != NULL; i++)
    if (clock == clocks[i].clock)
      return clocks[i].id;

  return NULL;
}

static enum Clock_identifiers get_clock(char *id)
{
  if (id != NULL)
  {
    for (int i = 0; clocks[i].id != NULL; i++)
      if (!strcmp(id, clocks[i].id))
        return clocks[i].clock;
  }

  return Clock_none;
}

static const char list_clocks_separator[] = "--------------------------------";

#define sep(x) (list_clocks_separator + sizeof(list_clocks_separator) - 1 - x)

static void list_clocks(void *cpm)
{
  /* Read information from the clock and power management unit. */

  printf("%-10s  %-15s  %-7s  %-6s  %-12s  %-12s  %-24s  %-12s  %-3s\n"
         "%-10s  %-15s  %-7s  %-6s  %-12s  %-12s  %-24s  %-12s  %-3s\n",
         "Id", "Clock", "Type", "Source", "Source Clock", "Source Freq.",
         "Parameters", "Frequency", "On",
         sep(10), sep(15), sep(7), sep(6), sep(12), sep(12),
         sep(24), sep(12), sep(3));

  for (int i = 0; clocks[i].id != NULL; i++)
  {
    uint32_t parameters[4];
    char parameter_str[25] = {0};
    const char *source_id = clock_id(cpm_get_source_clock(cpm, clocks[i].clock));
    int num_parameters = cpm_get_parameters(cpm, clocks[i].clock, parameters);

    for (int p = 0, pos = 0; p < num_parameters; p++)
    {
      int result = sprintf(parameter_str + pos, "%-7d ", parameters[p]);
      if (result < 0)
        break;
      pos += result;
    }

    printf("%-10s  %-15s  %-7s  %-6d  %-12s  %-12lld  %-24s  %-12lld  %-3s\n",
      clocks[i].id,
      clocks[i].name,
      cpm_clock_type(cpm, clocks[i].clock),
      cpm_get_source(cpm, clocks[i].clock),
      source_id != NULL ? source_id : "...",
      cpm_get_source_frequency(cpm, clocks[i].clock),
      parameter_str,
      cpm_get_frequency(cpm, clocks[i].clock),
      cpm_have_clock(cpm, clocks[i].clock) ? "on" : "off");
  }
}

static enum Clock_identifiers get_clock_by_name(void)
{
  char *token;
  enum Clock_identifiers clock;

  if ((token = read_token("Clock")) == NULL)
    return Clock_none;

  clock = get_clock(token);

  if (clock == Clock_none)
  {
    printf("Unrecognised clock: %s\n", token);
    return Clock_none;
  }

  return clock;
}

static void set_frequency(void *cpm)
{
  enum Clock_identifiers clock;
  unsigned int frequency;

  clock = get_clock_by_name();

  if (clock == Clock_none)
    return;

  if (!read_number("Frequency", &frequency))
    return;

  if (!cpm_set_frequency(cpm, clock, frequency))
    printf("Frequency not set.\n");
}

static void set_parameters(void *cpm)
{
  char *token;
  enum Clock_identifiers clock;
  uint32_t parameters[4];
  int num_parameters;

  clock = get_clock_by_name();

  if (clock == Clock_none)
    return;

  for (num_parameters = 0; num_parameters < 4; num_parameters++)
  {
    if ((token = read_token(NULL)) == NULL)
      break;

    parameters[num_parameters] = atoi(token);
  }

  if (!num_parameters)
  {
    printf("Parameters?\n");
    return;
  }

  if (!cpm_set_parameters(cpm, clock, num_parameters, parameters))
    printf("Too few or invalid parameters.\n");
}

static void set_source(void *cpm)
{
  enum Clock_identifiers clock;
  unsigned int source;

  clock = get_clock_by_name();

  if (clock == Clock_none)
    return;

  if (!read_number("Source", &source))
    return;

  cpm_set_source(cpm, clock, source);
}

static void start_clock(void *cpm)
{
  enum Clock_identifiers clock;

  clock = get_clock_by_name();

  if (clock == Clock_none)
    return;

  cpm_start_clock(cpm, clock);
}

static void stop_clock(void *cpm)
{
  enum Clock_identifiers clock;

  clock = get_clock_by_name();

  if (clock == Clock_none)
    return;

  cpm_stop_clock(cpm, clock);
}



/* DMA configuration. */

static l4_uint32_t dma_irq_start = 0, dma_irq_end = 0;

/* DMA operations. */

static int init_dma(void)
{
  unsigned int i;

  for (i = 0; i < num_dma_regions; i++)
    dma_regions[i].mem = L4_INVALID_CAP;

  /* Here, only one IRQ is used. */

  dma_irq = l4re_util_cap_alloc();
  return init_irq(0, dma_irq, dma_irq_start, dma_irq_end);
}

static void list_dma_regions(void)
{
  unsigned int num;
  struct dma_region *region;
  unsigned int i;

  for (num = 0; num < num_dma_regions; num++)
  {
    printf("Region %d: ", num);

    region = &dma_regions[num];

    if (l4_is_invalid_cap(region->mem))
      printf("(inactive)\n");
    else
    {
      printf("size = %d; align = %d; virtual = 0x%lx; physical = 0x%llx\ndata =",
             region->size, region->align, region->vaddr, region->paddr);

      for (i = 0; (i < region->size) && (i < 16); i++)
        printf(" %02x", *((uint8_t *) region->vaddr + i));

      printf("\n");
    }
  }
}

static void new_dma_channel(void *dma)
{
  int num = get_channel_number(num_dma_channels);

  if (num < 0)
    return;

  if (dma_channels[num] != NULL)
  {
    printf("Channel already defined.\n");
    return;
  }

  dma_channels[num] = dma_get_channel(dma, num, dma_irq);
}

static void new_dma_region(void)
{
  struct dma_region *region = _get_dma_region();

  if (region == NULL)
  {
    list_dma_regions();
    return;
  }

  if (l4_is_valid_cap(region->mem))
  {
    printf("Region already defined.\n");
    return;
  }

  if (!read_number("Size", &region->size))
    return;

  if (!read_number("Alignment", &region->align))
    return;

  if (get_dma_region(region->size, region->align, &region->vaddr, &region->paddr, &region->mem))
    printf("Could not allocate region.\n");
}

static void set_dma_region(void)
{
  char *token;
  struct dma_region *region = _get_dma_region();
  FILE *fp;

  if (region == NULL)
  {
    list_dma_regions();
    return;
  }

  if (l4_is_invalid_cap(region->mem))
  {
    printf("Region needs creating.\n");
    return;
  }

  memset((void *) region->vaddr, 0, region->size);

  if ((token = read_token("Filename")) == NULL)
    return;

  /* Populate the region from the file. */

  fp = fopen(token, "r");

  if (fp == NULL)
  {
    printf("File not readable.\n");
    return;
  }

  fread((char *) region->vaddr, sizeof(char), region->size, fp);
  fclose(fp);

  l4_cache_flush_data(region->vaddr, region->vaddr + region->size);
}

static void dma_transfer_data(void)
{
  uint32_t count, to_transfer, transferred;
  unsigned int source_address, destination_address, source_width, destination_width,
               transfer_unit_size, request_type;
  int source_increment, destination_increment;
  void *channel = get_channel(num_dma_channels, dma_channels, NULL);
  unsigned int i;

  if (channel == NULL)
    return;

  if (!read_number("Source address", &source_address))
    return;

  if (!read_number("Destination address", &destination_address))
    return;

  if (!read_number("Transfer size", &count))
    return;

  if (!read_flag("Source increment", &source_increment, "i", 1))
    return;

  if (!read_flag("Destination increment", &destination_increment, "i", 1))
    return;

  if (!read_number("Source width", &source_width))
    return;

  if (!read_number("Destination width", &destination_width))
    return;

  if (!read_number("Transfer unit size", &transfer_unit_size))
    return;

  if (!read_number("Request type", &request_type))
    return;

  /* Invalidate DMA region contents in case of transfers between regions. */

  for (i = 0; i < num_dma_regions; i++)
    l4_cache_inv_data(dma_regions[i].vaddr, dma_regions[i].vaddr + dma_regions[i].size);

  to_transfer = dma_transfer(channel, source_address, destination_address,
                             count,
                             source_increment, destination_increment,
                             source_width, destination_width,
                             transfer_unit_size,
                             request_type);

  transferred = to_transfer ? count - dma_wait(channel) : 0;

  printf("Transferred: %d\n", transferred);
}



/* GPIO operations. */

static char get_gpio_dir_label(unsigned int value, unsigned int level)
{
  return value == Fix_input ? (level ? 'I' : 'i') : (level ? 'O' : 'o');
}

static char get_gpio_irq_label(unsigned int value)
{
  switch (value)
  {
    case L4_IRQ_F_NEG_EDGE: return 'f';
    case L4_IRQ_F_POS_EDGE: return 'r';
    case L4_IRQ_F_LEVEL_LOW: return 'l';
    case L4_IRQ_F_LEVEL_HIGH: return 'h';
    default: return '?';
  }
}

static void list_gpios(void *gpio[])
{
  unsigned int port, pin;
  unsigned int func, value, level;

  printf("gpio: I = input high; i = input low; O = output high; o = output low\n"
         "irq:  h = high level; l = low level; r = rising edge; f = falling edge\n"
         "alt: function number\n\n");

  /* Show pin numbering. */

  printf("Port/Pin ");

  for (pin = 0; pin < 32; pin++)
  {
    if (!(pin % 10))
      printf(" %d", pin);
    else
      printf(" %d", pin % 10);
  }

  printf("\n");

  for (port = 0; port < num_gpio_ports; port++)
  {
    /* Show port and pin configuration. */

    printf("%c        ", gpio_port_labels[port]);

    for (pin = 0; pin < 32; pin++)
    {
      /* Pad below the first pin digit for multiples of ten other than zero. */

      if (pin && !(pin % 10))
        printf(" ");

      gpio_config_pad_get(gpio[port], pin, &func, &value);

      switch (func)
      {
        case Function_alt: printf(" %d", value); break;
        case Function_irq: printf(" %c", get_gpio_irq_label(value)); break;
        case Function_gpio:
        {
          level = gpio_get(gpio[port], pin);
          printf(" %c", get_gpio_dir_label(value, level));
          break;
        }
        default: printf(" ?"); break;
      }
    }

    printf("\n");
  }
}

static int get_port_and_pin(unsigned int *port, unsigned int *pin)
{
  char *token;

  if ((token = read_token("Port")) == NULL)
    return 0;

  if ((token[0] < 'A') || ((unsigned int) token[0] - (unsigned int) 'A' >= num_gpio_ports))
  {
    printf("Bad port: %c\n", token[0]);
    return 0;
  }

  *port = (unsigned int) token[0] - (unsigned int) 'A';

  if (!read_number("Pin", pin))
    return 0;

  return 1;
}

static void set_gpio_alt_func(void *gpio[])
{
  unsigned int port, pin, value;

  if (!get_port_and_pin(&port, &pin))
    return;

  if (!read_number("Function", &value))
    return;

  gpio_config_pad(gpio[port], pin, Function_alt, value);
}

static void set_gpio_pad(void *gpio[])
{
  char *token;
  unsigned int port, pin, mode, value = 0;

  if (!get_port_and_pin(&port, &pin))
    return;

  if ((token = read_token("Mode")) == NULL)
    return;

  if (!strcmp(token, "in"))
    mode = Fix_input;
  else if (!strcmp(token, "irq"))
    mode = Fix_irq;
  else if (!strcmp(token, "out"))
  {
    mode = Fix_output;

    if ((token = read_token(NULL)) != NULL)
      value = atoi(token);
  }
  else
  {
    printf("Mode not recognised.\n");
    return;
  }

  gpio_setup(gpio[port], pin, mode, value);
}

static void set_gpio_pull(void *gpio[])
{
  char *token;
  unsigned int port, pin, mode;

  if (!get_port_and_pin(&port, &pin))
    return;

  if ((token = read_token("Mode")) == NULL)
    return;

  if (!strcmp(token, "down"))
    mode = Pull_down;
  else if (!strcmp(token, "none"))
    mode = Pull_none;
  else if (!strcmp(token, "up"))
    mode = Pull_up;
  else
  {
    printf("Mode not recognised.\n");
    return;
  }

  gpio_config_pull(gpio[port], pin, mode);
}



/* I2C configuration. */

static l4_uint32_t i2c_irq_start = 0, i2c_irq_end = 0;

/* I2C operations. */

static int _i2c_read(void *channel, uint8_t *buf, unsigned length,
                     int stop, l4_cap_idx_t irqcap, unsigned timeout)
{
  l4_msgtag_t tag;

  i2c_start_read(channel, buf, length, stop);

  while (!i2c_read_done(channel))
  {
    tag = l4_irq_receive(irqcap, l4_timeout(L4_IPC_TIMEOUT_NEVER, l4_timeout_from_us(timeout)));

    // NOTE: Error not returned.

    if (l4_ipc_error(tag, l4_utcb()))
      break;

    if (i2c_failed(channel))
      break;

    i2c_read(channel);
  }

  if (stop)
    i2c_stop(channel);

  return i2c_have_read(channel);
}

static int _i2c_write(void *channel, uint8_t *buf, unsigned length,
                      int stop, l4_cap_idx_t irqcap, unsigned timeout)
{
  l4_msgtag_t tag;

  i2c_start_write(channel, buf, length, stop);

  while (!i2c_write_done(channel))
  {
    tag = l4_irq_receive(irqcap, l4_timeout(L4_IPC_TIMEOUT_NEVER, l4_timeout_from_us(timeout)));

    // NOTE: Error not returned.

    if (l4_ipc_error(tag, l4_utcb()))
      break;

    if (i2c_failed(channel))
      break;

    i2c_write(channel);
  }

  if (stop)
    i2c_stop(channel);

  return i2c_have_written(channel);
}

static void list_i2c_channels(void)
{
  unsigned int num;
  void *channel;

  printf("i2c: channel\n\n");

  for (num = 0; num < num_i2c_channels; num++)
  {
    printf("Channel %d: ", num);

    channel = i2c_channels[num];

    if (channel == NULL)
      printf("(inactive)\n");
    else
      printf("%d Hz\n", i2c_get_frequency(channel));
  }
}

static void new_i2c_channel(void *i2c)
{
  l4_cap_idx_t irqcap;
  int num = get_channel_number(num_i2c_channels);

  if (num < 0)
    return;

  irqcap = l4re_util_cap_alloc();

  if (init_irq(num, irqcap, i2c_irq_start, i2c_irq_end))
    return;

  i2c_channels[num] = i2c_get_channel(i2c, num);
  i2c_irqs[num] = irqcap;
}

static void i2c_get(void)
{
  void *channel;
  int num;
  uint8_t buffer[32];
  unsigned int address, reg, count;
  int i, transferred;

  channel = get_channel(num_i2c_channels, i2c_channels, &num);

  if (channel == NULL)
    return;

  if (!read_encoded_number("Address", "%2x", &address))
    return;

  if (address >= 0x80)
  {
    printf("Address must be less than 80.\n");
    return;
  }

  if (!read_encoded_number("Register", "%2x", &reg))
    return;

  if (!read_number(NULL, &count))
    count = 1;

  buffer[0] = (uint8_t) (reg & 0xff);

  i2c_set_target(channel, address);

  if (!_i2c_write(channel, buffer, 1, 0, i2c_irqs[num], 1000000))
  {
    printf("Register request failed.\n");
    return;
  }

  transferred = _i2c_read(channel, buffer, count, 1, i2c_irqs[num], 1000000);

  if (!transferred)
  {
    printf("Register read failed.\n");
    return;
  }

  for (i = 0; i < transferred; i++)
    printf("%02x ", buffer[i]);
  printf("\n");
}

static void i2c_scan(void)
{
  void *channel;
  unsigned int address;
  uint8_t buffer[1];
  int num;

  channel = get_channel(num_i2c_channels, i2c_channels, &num);

  if (channel == NULL)
    return;

  for (address = 0; address < 0x20; address++)
    printf("%02x ", address);
  printf("\n");

  for (address = 0; address < 0x20; address++)
    printf("-- ");

  for (address = 0; address < 0x80; address++)
  {
    if ((address % 32) == 0)
      printf("\n");

    i2c_set_target(channel, address);

    if (_i2c_read(channel, buffer, 1, 1, i2c_irqs[num], 1000))
      printf("%02x ", address);
    else
      printf("-- ");
  }

  printf("\n");
  for (address = 0; address < 0x20; address++)
    printf("-- ");
  printf("\n\n");
}



/* Memory operations. */

static void list_memory_regions(void)
{
  unsigned int num;
  struct memory_region *region;
  unsigned int i;

  for (num = 0; num < num_memory_regions; num++)
  {
    printf("Region %d: ", num);

    region = &memory_regions[num];

    if (!region->addr)
      printf("(inactive)\n");
    else
    {
      printf("size = %d; virtual = 0x%lx; data =",
             region->size, region->addr);

      for (i = 0; (i < region->size) && (i < 16); i++)
        printf(" %02x", *((uint8_t *) region->addr + i));

      printf("\n");
    }
  }
}

static void new_memory_region(void)
{
  struct memory_region *region = _get_memory_region();

  if (region == NULL)
  {
    list_memory_regions();
    return;
  }

  if (region->addr)
  {
    printf("Region already defined.\n");
    return;
  }

  if (!read_number("Size", &region->size))
    return;

  if (get_memory_region(region->size, &region->addr))
    printf("Could not allocate region.\n");
}

static void set_memory_region(void)
{
  char *token;
  struct memory_region *region = _get_memory_region();
  FILE *fp;

  if (region == NULL)
  {
    list_memory_regions();
    return;
  }

  if (!region->addr)
  {
    printf("Region needs creating.\n");
    return;
  }

  memset((void *) region->addr, 0, region->size);

  if ((token = read_token("Filename")) == NULL)
    return;

  /* Populate the region from the file. */

  fp = fopen(token, "r");

  if (fp == NULL)
  {
    printf("File not readable.\n");
    return;
  }

  fread((char *) region->addr, sizeof(char), region->size, fp);
  fclose(fp);

  l4_cache_flush_data(region->addr, region->addr + region->size);
}



/* RTC operations. */

static void rtc_reset(void *rtc, void *cpm)
{
  unsigned int seconds;

  if (!read_number("Seconds", &seconds))
    return;

  /* NOTE: Assuming EXCLK/512 as RTC source. */

  uint32_t rtc_seconds = rtc_get_seconds(rtc);
  uint32_t value = seconds * cpm_get_frequency(cpm, Clock_external) / 512;

  rtc_alarm_disable(rtc);
  rtc_set_alarm_seconds(rtc, rtc_seconds + value);
  rtc_hibernate(rtc);
}

static void _rtc_set_seconds(void *rtc, int alarm)
{
  unsigned int seconds;

  if (!read_number("Seconds", &seconds))
    return;

  (alarm ? rtc_set_alarm_seconds : rtc_set_seconds)(rtc, seconds);
}



/* SPI operations. */

static void new_spi_channel(void *spi, void *gpio[])
{
  unsigned int control_port, control_pin, control_alt_func_input;
  int control_alt_func;
  void *control_chip;
  int num = get_channel_number(num_spi_channels);
  void *channel;
  uint32_t frequency;

  if (num < 0)
    return;

  channel = get_channel(num_dma_channels, dma_channels, NULL);

  if (channel == NULL)
    return;

  if (!read_number("Frequency", &frequency))
    return;

  if (!get_port_and_pin(&control_port, &control_pin))
    control_chip = NULL;
  else
    control_chip = gpio[control_port];

  if (!read_number("Function", &control_alt_func_input))
    control_alt_func = -1;
  else
    control_alt_func = control_alt_func_input;

  spi_channels[num] = spi_get_channel(spi, num, channel, frequency,
                                      control_chip, control_pin,
                                      control_alt_func);
}

static void new_spi_channel_gpio(void *gpio[])
{
  unsigned int clock_port, clock_pin, data_port, data_pin, enable_port,
               enable_pin, control_port, control_pin;
  void *control_chip;
  int num = get_channel_number(num_spi_channels);
  uint32_t frequency;

  if (num < 0)
    return;

  if (!read_number("Frequency", &frequency))
    return;

  if (!get_port_and_pin(&clock_port, &clock_pin))
    return;

  if (!get_port_and_pin(&data_port, &data_pin))
    return;

  if (!get_port_and_pin(&enable_port, &enable_pin))
    return;

  if (!get_port_and_pin(&control_port, &control_pin))
    control_chip = NULL;
  else
    control_chip = gpio[control_port];

  spi_channels[num] = spi_get_channel_gpio(frequency,
                                           gpio[clock_port], clock_pin,
                                           gpio[data_port], data_pin,
                                           gpio[enable_port], enable_pin,
                                           control_chip, control_pin);
}

static void spi_control(int acquire)
{
  unsigned int level;
  void *channel = get_channel(num_spi_channels, spi_channels, NULL);

  if (acquire)
  {
    if (!read_number("Level", &level))
      return;

    spi_acquire_control(channel, level);
  }
  else
    spi_release_control(channel);
}

static void spi_send_data(void)
{
  void *channel = get_channel(num_spi_channels, spi_channels, NULL);
  int bytes;
  unsigned int byte;
  uint8_t buffer[256];

  bytes = 0;

  while (read_encoded_number(NULL, "%2x", &byte))
  {
    buffer[bytes] = (uint8_t) (byte & 0xff);
    bytes++;
  }

  spi_send(channel, bytes, buffer);
}

static void spi_send_data_units(void)
{
  void *channel = get_channel(num_spi_channels, spi_channels, NULL);
  unsigned int char_size, unit_size, value;
  uint8_t buffer[256];
  int byte = 0;

  if (!read_number("Unit size", &unit_size))
    return;

  if (!read_number("Character size", &char_size))
    return;

  /* Read hex digits for bytes. Multiple bytes make up each unit and are read
     from most to least significant. Where the unit size exceeds the character
     size, the last bit before the character indicates the GPC bit. */

  while ((byte < 256) && read_encoded_number(NULL, "%2x", &value))
    buffer[byte++] = value;

  /* Explicitly indicate big endian data. */

  spi_send_units(channel, byte, (uint8_t *) buffer, unit_size, char_size, 1);
}

static void spi_transfer_data(void)
{
  void *channel = get_channel(num_spi_channels, spi_channels, NULL);
  struct dma_region *dma_region = NULL, *desc_region;
  struct memory_region *memory_region = NULL;
  unsigned int char_size, unit_size;
  uint32_t count, transferred;
  int using_dma;

  if (channel == NULL)
    return;

  if (!read_flag("Using DMA", &using_dma, "d", 1))
    return;

  if (using_dma)
    dma_region = _get_dma_region();
  else
    memory_region = _get_memory_region();

  if ((dma_region == NULL) && (memory_region == NULL))
    return;

  if (!read_number("Transfer size", &count))
    return;

  if (!read_number("Unit size", &unit_size))
    return;

  if (!read_number("Character size", &char_size))
    return;

  if (using_dma)
  {
    desc_region = _get_dma_region();

    transferred = spi_transfer(channel, dma_region->vaddr, dma_region->paddr, count,
                               unit_size, char_size,
                               desc_region != NULL ? desc_region->vaddr : 0,
                               desc_region != NULL ? desc_region->paddr : 0);
  }
  else
    transferred = spi_transfer(channel, memory_region->addr, 0, count,
                               unit_size, char_size, 0, 0);

  printf("Transferred: %d\n", transferred);
}



/* Command processing. */

static void handle_aic(void *aic)
{
  char *token;

  if ((token = read_token(NULL)) != NULL)
  {
    if (!strcmp(token, "l") || !strcmp(token, "list"))
      list_channels(num_aic_channels, aic_channels);
    else if (!strcmp(token, "c") || !strcmp(token, "channel"))
      new_aic_channel(aic);
    else if (!strcmp(token, "t") || !strcmp(token, "transfer"))
      aic_transfer_data();
    else
      printf("aic channel | list | transfer\n");
  }
  else
    list_channels(num_aic_channels, aic_channels);
}

static void handle_cpm(void *cpm)
{
  char *token;

  if ((token = read_token(NULL)) != NULL)
  {
    if (!strcmp(token, "l") || !strcmp(token, "list"))
      list_clocks(cpm);
    else if (!strcmp(token, "f") || !strcmp(token, "frequency"))
      set_frequency(cpm);
    else if (!strcmp(token, "p") || !strcmp(token, "parameters"))
      set_parameters(cpm);
    else if (!strcmp(token, "s") || !strcmp(token, "source"))
      set_source(cpm);
    else if (!strcmp(token, "start"))
      start_clock(cpm);
    else if (!strcmp(token, "stop"))
      stop_clock(cpm);
    else
      printf("cpm list | frequency | parameters | source | start | stop\n");
  }
  else
    list_clocks(cpm);
}

static void handle_dma(void *dma)
{
  char *token;

  if ((token = read_token(NULL)) != NULL)
  {
    if (!strcmp(token, "l") || !strcmp(token, "list"))
      list_channels(num_dma_channels, dma_channels);
    else if (!strcmp(token, "c") || !strcmp(token, "channel"))
      new_dma_channel(dma);
    else if (!strcmp(token, "r") || !strcmp(token, "region"))
      new_dma_region();
    else if (!strcmp(token, "s") || !strcmp(token, "set"))
      set_dma_region();
    else if (!strcmp(token, "t") || !strcmp(token, "transfer"))
      dma_transfer_data();
    else
      printf("dma channel | list | region | set | transfer\n");
  }
  else
    list_channels(num_dma_channels, dma_channels);
}

static void handle_gpio(void *gpio[])
{
  char *token;

  if ((token = read_token(NULL)) != NULL)
  {
    if (!strcmp(token, "l") || !strcmp(token, "list"))
      list_gpios(gpio);
    else if (!strcmp(token, "a") || !strcmp(token, "alt"))
      set_gpio_alt_func(gpio);
    else if (!strcmp(token, "i") || !strcmp(token, "io"))
      set_gpio_pad(gpio);
    else if (!strcmp(token, "p") || !strcmp(token, "pull"))
      set_gpio_pull(gpio);
    else
      printf("gpio list | alt | io | pull\n");
  }
  else
    list_gpios(gpio);
}

static void handle_i2c(void *i2c)
{
  char *token;

  if ((token = read_token(NULL)) != NULL)
  {
    if (!strcmp(token, "l") || !strcmp(token, "list"))
      list_i2c_channels();
    else if (!strcmp(token, "c") || !strcmp(token, "channel"))
      new_i2c_channel(i2c);
    else if (!strcmp(token, "g") || !strcmp(token, "get"))
      i2c_get();
    else if (!strcmp(token, "s") || !strcmp(token, "scan"))
      i2c_scan();
    else
      printf("i2c channel | get | list | scan\n");
  }
  else
    list_i2c_channels();
}

static void handle_memory(void)
{
  char *token;

  if ((token = read_token(NULL)) != NULL)
  {
    if (!strcmp(token, "r") || !strcmp(token, "region"))
      new_memory_region();
    else if (!strcmp(token, "s") || !strcmp(token, "set"))
      set_memory_region();
    else
      printf("memory region | set\n");
  }
  else
    list_memory_regions();
}

static void handle_rtc(void *rtc, void *cpm)
{
  char *token;

  if ((token = read_token(NULL)) != NULL)
  {
    if (!strcmp(token, "d") || !strcmp(token, "disable"))
      rtc_disable(rtc);
    else if (!strcmp(token, "e") || !strcmp(token, "enable"))
      rtc_enable(rtc);
    else if (!strcmp(token, "g") || !strcmp(token, "get"))
      printf("seconds = %d\n", rtc_get_seconds(rtc));
    else if (!strcmp(token, "ga") || !strcmp(token, "get-alarm"))
      printf("seconds = %d\n", rtc_get_alarm_seconds(rtc));
    else if (!strcmp(token, "p") || !strcmp(token, "power-down"))
      rtc_power_down(rtc);
    else if (!strcmp(token, "r") || !strcmp(token, "reset"))
      rtc_reset(rtc, cpm);
    else if (!strcmp(token, "s") || !strcmp(token, "set"))
      _rtc_set_seconds(rtc, 0);
    else if (!strcmp(token, "sa") || !strcmp(token, "set-alarm"))
      _rtc_set_seconds(rtc, 1);
    else
      printf("rtc disable | enable | get | get-alarm | power-down | reset | set | set-alarm\n");
  }
  else
    printf("rtc disable | enable | get | get-alarm | power-down | reset | set | set-alarm\n");
}

static void handle_spi(void *spi, void *gpio[])
{
  char *token;

  if ((token = read_token(NULL)) != NULL)
  {
    if (!strcmp(token, "l") || !strcmp(token, "list"))
      list_channels(num_spi_channels, spi_channels);
    else if (!strcmp(token, "a") || !strcmp(token, "control-acquire"))
      spi_control(1);
    else if (!strcmp(token, "r") || !strcmp(token, "control-release"))
      spi_control(0);
    else if (!strcmp(token, "c") || !strcmp(token, "channel"))
      new_spi_channel(spi, gpio);
    else if (!strcmp(token, "g") || !strcmp(token, "gpio") || !strcmp(token, "gpio-channel"))
      new_spi_channel_gpio(gpio);
    else if (!strcmp(token, "s") || !strcmp(token, "send"))
      spi_send_data();
    else if (!strcmp(token, "S") || !strcmp(token, "send-units"))
      spi_send_data_units();
    else if (!strcmp(token, "t") || !strcmp(token, "transfer"))
      spi_transfer_data();
    else
      printf("spi channel | control-acquire | control-release | gpio-channel | list | send | send-units | transfer\n");
  }
  else
    list_channels(num_spi_channels, spi_channels);
}



int main(void)
{
  l4_addr_t aic_base = 0, aic_base_end = 0;
  l4_addr_t aic_phys_base = 0, aic_phys_base_end = 0;
  l4_addr_t cpm_base = 0, cpm_base_end = 0;
  l4_addr_t dma_base = 0, dma_base_end = 0;
  l4_addr_t gpio_base = 0, gpio_base_end = 0;
  l4_addr_t i2c_base = 0, i2c_base_end = 0;
  l4_addr_t rtc_base = 0, rtc_base_end = 0;
  l4_addr_t ssi_base = 0, ssi_base_end = 0;
  l4_addr_t ssi_phys_base = 0, ssi_phys_base_end = 0;
  void *aic, *cpm, *dma, *gpio[num_gpio_ports], *i2c, *rtc, *spi;
  int result = 0;
  unsigned int port;

  icucap = l4re_env_get_cap("icu");

  /* Obtain resource details describing I/O memory. */

  printf("Access CPM...\n");

  if ((result = get_memory(io_memory_regions[CPM], &cpm_base, &cpm_base_end)) < 0)
    return 1;

  printf("CPM at 0x%lx...0x%lx.\n", cpm_base, cpm_base_end);

  cpm = cpm_init(cpm_base);

  printf("Access DMA...\n");

  if ((result = get_memory(io_memory_regions[DMA], &dma_base, &dma_base_end)) < 0)
    return 1;

  printf("DMA at 0x%lx...0x%lx.\n", dma_base, dma_base_end);

  dma = dma_init(dma_base, dma_base_end, cpm);

  if (get_irq(io_memory_regions[DMA], &dma_irq_start, &dma_irq_end) < 0)
    return 1;

  printf("IRQ range at %d...%d.\n", dma_irq_start, dma_irq_end);

  if (init_dma())
    return 1;

  dma_enable(dma);

  printf("Access GPIO...\n");

  if ((result = get_memory(io_memory_regions[GPIO], &gpio_base, &gpio_base_end)) < 0)
    return 1;

  printf("GPIO at 0x%lx...0x%lx.\n", gpio_base, gpio_base_end);

  for (port = 0; port < num_gpio_ports; port++)
    gpio[port] = gpio_init(gpio_base + port * 0x100, gpio_base + (port + 1) * 0x100,
                           32, gpio_ports[port].pull_ups, gpio_ports[port].pull_downs);

  printf("Access I2C...\n");

  if ((result = get_memory(io_memory_regions[I2C], &i2c_base, &i2c_base_end)) < 0)
    return 1;

  printf("I2C at 0x%lx...0x%lx.\n", i2c_base, i2c_base_end);

  i2c = i2c_init(i2c_base, i2c_base_end, cpm, 100000);

  if (get_irq(io_memory_regions[I2C], &i2c_irq_start, &i2c_irq_end) < 0)
    return 1;

  printf("IRQ range at %d...%d.\n", i2c_irq_start, i2c_irq_end);

  printf("Access AIC...\n");

  if ((result = get_memory_complete(io_memory_regions[AIC], &aic_base, &aic_base_end,
                                    &aic_phys_base, &aic_phys_base_end)) < 0)
    return 1;

  printf("AIC at 0x%lx...0x%lx.\n", aic_base, aic_base_end);

  aic = aic_init(aic_phys_base, aic_base, aic_base_end, cpm);

  printf("Access RTC...\n");

  if ((result = get_memory(io_memory_regions[RTC], &rtc_base, &rtc_base_end)) < 0)
    return 1;

  printf("RTC at 0x%lx...0x%lx.\n", rtc_base, rtc_base_end);

  rtc = rtc_init(rtc_base, cpm);

  printf("Access SSI...\n");

  if ((result = get_memory_complete(io_memory_regions[SSI], &ssi_base, &ssi_base_end,
                                    &ssi_phys_base, &ssi_phys_base_end)) < 0)
    return 1;

  printf("SSI at 0x%lx...0x%lx.\n", ssi_base, ssi_base_end);

  spi = spi_init(ssi_phys_base, ssi_base, ssi_base_end, cpm);

  /* Start the interactive session. */

  printf("aic, cpm, dma, gpio, i2c, rtc, spi\n");

  while (1)
  {
    char cmdline[256], *token;

    printf("> ");

    token = fgets(cmdline, 256, stdin);

    if (token == NULL)
      break;

    if ((token = strtok(cmdline, delimiters)) == NULL)
      continue;

    /* AIC/I2S commands. */

    if (!strcmp(token, "a") || !strcmp(token, "aic") || !strcmp(token, "i2s"))
      handle_aic(aic);

    /* CPM commands. */

    else if (!strcmp(token, "c") || !strcmp(token, "cpm"))
      handle_cpm(cpm);

    /* DMA commands. */

    else if (!strcmp(token, "d") || !strcmp(token, "dma"))
      handle_dma(dma);

    /* GPIO commands. */

    else if (!strcmp(token, "g") || !strcmp(token, "gpio"))
      handle_gpio(gpio);

    /* I2C commands. */

    else if (!strcmp(token, "i") || !strcmp(token, "i2c"))
      handle_i2c(i2c);

    /* Generic memory commands. */

    else if (!strcmp(token, "m") || !strcmp(token, "mem") || !strcmp(token, "memory"))
      handle_memory();

    /* RTC commands. */

    else if (!strcmp(token, "r") || !strcmp(token, "rtc"))
      handle_rtc(rtc, cpm);

    /* SPI commands. */

    else if (!strcmp(token, "s") || !strcmp(token, "spi"))
      handle_spi(spi, gpio);

    /* Comments and blank lines. */

    else if (strncmp(token, "#", 1) && strlen(token))
      printf("Command?\n");
  }

  printf("End of session.\n");
  return 0;
}
